// $Id: Preflight.java 22021 2015-09-25 14:28:08Z mike $

import java.awt.Color;
import java.awt.color.*;
import java.io.*;
import java.util.*;
import org.faceless.pdf2.*;

/** 
 * An example demonstrating how to "preflight" a document - in this case, to convert it
 * to PDF/A or PDF/X. This new example (first supplied with 2.18) primarily shows the
 * OutputProfiler class and how to use it.
 */
public class Preflight {

    public static void main(String[] args) throws Exception {
        OutputProfiler.AutoEmbeddingFontAction fontaction = new OutputProfiler.AutoEmbeddingFontAction();

        float hairline = 0.1f;                          // Width to replace "hairlines"
        OutputProfile target = null;
        ICC_Profile icc = null;
        String outfile = "preflight.pdf";

        for (int i=0;i<args.length;i++) {
            if (args[i].equals("--font")) {
                InputStream in = new FileInputStream(args[++i]);
                PDFFont font = new OpenTypeFont(in, 2);
                in.close();
                fontaction.add(font);
                System.out.println("* Added font "+args[i]);
            } else if (args[i].equals("--fontdir")) {
                File dir = new File(args[++i]);
                File[] files = dir.listFiles();
                for (int j=0;j<files.length;j++) {
                    if (files[j].getName().endsWith(".ttf") || files[j].getName().endsWith(".otf")) {
                        try {
                            InputStream in = new FileInputStream(files[j]);
                            PDFFont font = new OpenTypeFont(in, 2);
                            in.close();
                            fontaction.add(font);
                            System.out.println("* Added font \""+font.getBaseName()+"\" from \""+files[j]+"\"");
                        } catch (Exception e) {
                        }
                    }
                }
            } else if (args[i].equals("--pdfa1b")) {
                target = OutputProfile.PDFA1b_2005;
            } else if (args[i].equals("--pdfa2b")) {
                target = OutputProfile.PDFA2b;
            } else if (args[i].equals("--pdfa3b")) {
                target = OutputProfile.PDFA3b;
            } else if (args[i].equals("--pdfx3")) {
                target = OutputProfile.PDFX3_2003;
            } else if (args[i].equals("--srgb")) {
                icc = ICC_Profile.getInstance(ColorSpace.CS_sRGB);
            } else if (args[i].equals("--icc")) {
                icc = ICC_Profile.getInstance(args[++i]);
            } else if (args[i].equals("--output")) {
                outfile = args[++i];
            } else if (args[i].equals("--hairline")) {
                hairline = Float.parseFloat(args[++i]);
            } else if (i != args.length-1) {
                usage();
            } else {
                if (target == null && icc == null) {
                    System.out.println("No profile specified, assuming PDFA-1b with sRGB output intent");
                    target = OutputProfile.PDFA1b_2005;
                    icc = ICC_Profile.getInstance(ColorSpace.CS_sRGB);
                } else if (target == null) {
                    System.out.println("No profile specified, assuming PDFA-1b");
                    target = new OutputProfile(OutputProfile.PDFA1b_2005);
                } else if (target == null) {
                    System.out.println("No ICC profile specified, assuming sRGB output intent");
                    icc = ICC_Profile.getInstance(ColorSpace.CS_sRGB);
                }

                // Load the PDF and see if it matches selected target profile
                PDF pdf = new PDF(new PDFReader(new File(args[i])));
                OutputProfiler profiler = new OutputProfiler(new PDFParser(pdf));
                OutputProfile profile = profiler.getProfile();
                OutputProfile.Feature[] mismatch = profile.isCompatibleWith(target);
                if (mismatch == null) {
                    System.out.println("PDF already matches "+target.getProfileName());
                } else {
                    System.out.println("PDF doesn't match "+target.getProfileName());
                    for (int j=0;j<mismatch.length;j++) {
                        if (profile.isSet(mismatch[j])) {
                            System.out.println("* \"" + mismatch[j] + "\" is set but disallowed");
                        } else {
                            System.out.println("* \"" + mismatch[j] + "\" is required but missing");
                        }
                    }
                    System.out.println("Repairing");

                    // Modify the PDF to match. We create a new OutputProfile based on the specified
                    // targer profile, set it's intent based on the ICC profile we've selected,
                    // add the font and color actions, then "apply" the profile. If this succeeds
                    // without exception, save the file.
                    target = new OutputProfile(target);
                    target.setOutputIntent(null, null, null, null, null, icc);
                    profiler.setFontAction(fontaction);
                    profiler.setColorAction(new OutputProfiler.ProcessColorAction(target));
                    profiler.setHairlineWidth(hairline);
                    profiler.apply(target);
                    System.out.println("Done. Writing to \""+outfile+"\"");
                    pdf.render(new FileOutputStream(outfile));
                }
            }
        }
    }

    private static void usage() {
        System.out.println("Usage: java Preflight [--pdfa1b | pdfa2b | pdfa3b | pdfx3] [--icc <profile>]");
        System.out.println("                      [--hairline <width>] [--font <fontfile>");
        System.out.println("                      [--fontdir <fontdir>] [--output <file> file.pdf");
        System.out.println();
        System.out.println("    --pdfa1b                   Try to verify or convert the PDF to be");
        System.out.println("                               compatible with PDF/A-1b:2005");
        System.out.println("    --pdfa2b                   Try to verify or convert the PDF to be");
        System.out.println("                               compatible with PDF/A-2b:2011");
        System.out.println("    --pdfa3b                   Try to verify or convert the PDF to be");
        System.out.println("                               compatible with PDF/A-3b:2012");
        System.out.println("    --pdfx3                    Try to verify or convert the PDF to be");
        System.out.println("                               compatible with PDF/X-3:2002");
        System.out.println("    --pdfx3                    Try to verify or convert the PDF to be");
        System.out.println("                               compatible with PDF/X-3:2002");
        System.out.println();
        System.out.println("    --icc <profile>            Specify the ICC profile to use as the output-");
        System.out.println("                               intent when converting to onve of the above");
        System.out.println("                               profiles. Should be the path to an ICC profile");
        System.out.println("                               file or the string \"sRGB\"");
        System.out.println();
        System.out.println("    --font <fontfile>          Add a font to the list of fonts that will be");
        System.out.println("                               considered as candidates to replace any");
        System.out.println("                               unembedded fonts in the PDF. The font should");
        System.out.println("                               be a TrueType font");
        System.out.println();
        System.out.println("    --fontdir <dir>            Scan the directory for any files ending in");
        System.out.println("                               .ttf or .otf, and add those fonts to the list");
        System.out.println("                               of candidates for replacement (as above)");
        System.out.println();
        System.out.println("    --hairline <width>         Specify the width of the lines used to");
        System.out.println("                               replace any hairlines in the PDF");
        System.out.println();
        System.out.println("    --output <file>            Specify the file to write the modified PDF");
        System.out.println("                               to. The default is \"preflight.pdf\"");
        System.out.println();
        System.out.println("Preflight will scan the specified PDF file to see if it matches the specified");
        System.out.println("Output-Profile - if it doesn't, it will attempt to modify the file to make it");
        System.out.println("compatible. Typically this involves replacing unembedded fonts with embedded");
        System.out.println("equivalent, and converting device-dependent ones to those based on an ICC profile.");
        System.exit(1);
    }

}
