// $Id: StructuredPDF.java 24934 2017-03-01 17:28:42Z mike $

import java.util.*;
import java.io.*;
import java.awt.Color;
import org.faceless.pdf2.*;

/**
 * A worked example showing how to use the beginTag/endTag methods
 * to add structure to a PDF.
 *
 * This example was revised in 2.19 and the resulting file is now
 * PDF/UA compliant.
 */
public class StructuredPDF {

    public static void main(String[] args) throws IOException {
        PDF pdf = new PDF();

        // Follow steps are all required to make the PDF compliant with PDF/UA
        OutputProfile profile = new OutputProfile(OutputProfile.Default);
        profile.setRequired(OutputProfile.Feature.InfoMeetsPDFUA1);
        pdf.setOutputProfile(profile);
        pdf.setLocale(Locale.US);                               // Locale must be set
        pdf.setInfo("Title", "An Accessible PDF");              // Title must be set
        pdf.setOption("view.displayDocTitle", Boolean.TRUE);    // This property must be true
        pdf.setOption("marked", Boolean.TRUE);                  // ... and PDF must be "marked"

        PDFPage page = pdf.newPage("Letter");
        PDFCanvas canvas = new PDFCanvas(612, 792);
        int margin = 36;

        // PDF/UA requires embedded fonts.
        PDFStyle style = new PDFStyle();
        OpenTypeFont font = new OpenTypeFont(new FileInputStream("resources/fonts/bluehigh.ttf"), 2);
        style.setFont(font, 24);
        style.setFillColor(Color.black);

        PDFStyle linkstyle = new PDFStyle(style);
        linkstyle.setFillColor(Color.blue);
        linkstyle.setTextUnderline(false);
        // PDF/UA requires all vector graphics that aren't significant to be marked as
        // artifacts. Currently this isn't done for automatically-created content like
        // link underling. Turn it off for now.

        Locale spainlocale = new Locale("es", "ES");
        LayoutBox box = new LayoutBox(page.getWidth() - margin - margin);
        Map<String,Object> atts = new HashMap<String,Object>();

        //--------------------------------------------------------------
        // Here we create the tagged content. 
        //
        // On the left we have the commands for         Here on the right we have
        // the PDF API                                  the analagous HTML-style
        //                                              tags.
        //
        // Of course the HTML equivalents aren't exact - PDF structure tags are
        // not a direct match, but they should help to explain the structure
        //
        // It is quite possible to use the wrong tags and create a file that isn't
        // PDF/UA compliant as a result. Ensuring you meed these requirements is
        // currently up to you.
        //
        page.beginTag("Art", null);                     // <article>

        atts.clear();
        atts.put("id", "para1");
        box.beginTag("P", atts);                        // <p id="para1">

        box.addText("Hello, ", style, null);            // hello

        atts.clear();
        atts.put("id", "span1");
        atts.put("class", "username");
        box.beginTag("Span", atts);                     // <span id="span1" class="username">

        box.addText("Mr Smith", style, null);           // Mr Smith

        box.addLineBreak(style);                        // <br/>
        box.endTag();                                   // </span>
        box.endTag();                                   // </p>

        atts.clear();
        atts.put("id", "para2");
        atts.put("lang", spainlocale);
        box.beginTag("P", atts);                        // <p id="para2" lang="en-ES">

        box.addText("Hola, ", style, null);             // Hola

        atts.clear();
        atts.put("id", "span2");
        atts.put("class", "username");
        box.beginTag("Span", atts);                     // <span id="span2" class="username">

        AnnotationLink link = new AnnotationLink();
        link.setAction(PDFAction.goToURL("http://bfo.com"));
        atts.clear();
        atts.put("annotation", link);
        box.beginTag("Link", atts);                     // <a href="http://bfo.com">

        LayoutBox.Text text = box.addText("Se\u00f1or Smith", linkstyle, null);  // Señor Smith

        box.endTag();                                   // </a>
        box.endTag();                                   // </span>
        box.endTag();                                   // </p>
        box.flush();

        page.drawLayoutBox(box, margin, page.getHeight() - margin);

        // Add an "artifact" - a page number.
        // See http://www.w3.org/TR/2014/NOTE-WCAG20-TECHS-20140408/pdf.html#PDF4
        // Fully accessible documents should label any content that is not part of
        // the logical page content as an Artifact. This includes vector graphics!

        page.beginTag("Artifact", null);                // <artifact>
        box = new LayoutBox(100);
        style.setFont(font, 12);
        box.addText("Page 1", style, null);
        style.setFont(font, 24);
        page.drawLayoutBox(box, page.getWidth() / 2, page.getHeight() - 20);
        page.endTag();                                  // <artifact>

        page.endTag();                                  // </article>

        // Set the link annotation to match our hyperlink
        link.setCorners(text.getCorners(margin, page.getHeight() - margin));
        link.setContents("Link to the BFO home page");  // Description is required
        link.setPage(page);

        OutputStream fo = new FileOutputStream("StructuredPDF.pdf");
        pdf.render(fo);
        fo.close();
    }

}
