// $Id: EmbeddedFilePanel.java 21547 2015-06-19 11:10:33Z chris $

package org.faceless.pdf2.viewer2.feature;

import org.faceless.pdf2.viewer2.*;
import org.faceless.pdf2.*;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.io.InputStream;
import java.util.*;
import java.util.List;

/**
 * Create a {@link SidePanel} that will display a list of embedded files in the PDF.
 *
 * <span class="featurename">The name of this feature is <span class="featureactualname">ShowHideEmbeddedFiles</span></span>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.18
 */
public class EmbeddedFilePanel extends SidePanelFactory {

    /**
     * Create a new EmbeddedFilePanel
     */
    public EmbeddedFilePanel() {
        super("ShowHideEmbeddedFiles");
    }

    public boolean isSidePanelRequired(PDF pdf) {
        // If we are still loading, assume that we are required
        // until all pages have arrived and we can safely say we are not.
        return pdf.getLoadState(-1) != null || !pdf.getEmbeddedFiles().isEmpty();
    }

    public SidePanel createSidePanel() {
        return new EmbeddedFilePanelImpl(this);
    }
}

/**
 * A {@link SidePanel} representing the "Embedded Files" or "Portfolio" panel.
 *
 * <p><i>
 * This code is copyright the Big Faceless Organization. You're welcome to
 * use, modify and distribute it in any form in your own projects, provided
 * those projects continue to make use of the Big Faceless PDF library.
 * </i></p>
 */
class EmbeddedFilePanelImpl extends JPanel implements SidePanel, ListCellRenderer<EmbeddedFile>, PropertyChangeListener, Runnable {

    private EmbeddedFilePanel factory;
    private JScrollPane scrollpane;
    private DocumentPanel docpanel;
    private Icon icon;
    private List<EmbeddedFile> model;
    private List<ListDataListener> ldl = new ArrayList<ListDataListener>();
    private JList<EmbeddedFile> list;
    private JLabel spinner;

    EmbeddedFilePanelImpl(EmbeddedFilePanel factory) {
        super(new BorderLayout());
        this.icon = new ImageIcon(PDFViewer.class.getResource("resources/icons/attach.png"));
        setOpaque(true);
        this.factory = factory;
        model = new ArrayList<EmbeddedFile>();
        list = new JList<EmbeddedFile>(this.new Model());
        list.setCellRenderer(this);
        list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        list.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent event) {
                if (event.getClickCount() == 2) {
                    int ix = list.locationToIndex(event.getPoint());
                    if (ix >= 0) {
                        EmbeddedFile attachment = model.get(ix);
                        handleClick(attachment);
                    }
                }
            }
        });
        scrollpane = new JScrollPane();
        scrollpane.setViewportView(list);
        Icon spinnerIcon = new ImageIcon(PDFViewer.class.getResource("resources/spinner.gif"));
        spinner = new JLabel(spinnerIcon);
    }

    public Icon getIcon() {
        return icon;
    }

    public String getName() {
        return "EmbeddedFiles";
    }

    public void panelVisible() {
    }

    public void panelHidden() {
    }

    public void setDocumentPanel(DocumentPanel docpanel) {
        if (docpanel == this.docpanel) {
            return;
        }
        this.docpanel = docpanel;
        PDF pdf = docpanel == null ? null : docpanel.getPDF();
        if (pdf != null) {
            if (pdf.getLoadState(-1) != null) {
                boolean show = factory.getFeatureProperty(docpanel.getViewer(), "HideUntilKnown") == null;
                if (show) {
                    remove(scrollpane);
                    add(spinner, BorderLayout.CENTER);
                    doRevalidate();
                } else {
                    setVisible(false);
                }
                docpanel.getLinearizedSupport().invokeOnDocumentLoad(this);
            } else {
                run();
            }
        }
    }

    /**
     * Run on document load.
     */
    public void run() {
        PDF pdf = docpanel.getPDF();
        init(pdf);
        pdf.addPropertyChangeListener(this);
    }

    private void init(PDF pdf) {
        Map<String,EmbeddedFile> attachments = pdf.getEmbeddedFiles();
        if (attachments.isEmpty()) {
            // All pages now loaded and there are no attachments
            docpanel.removeSidePanel(this);
        } else {
            model.clear();
            model.addAll(attachments.values());
            fireContentsChanged();
            remove(spinner);
            add(scrollpane, BorderLayout.CENTER);
            setVisible(true);
            docpanel.addSidePanel(this);
            doRevalidate();
        }
    }

    private void doRevalidate() {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                revalidate();
            }
        });
    }

    /**
     * Document changed.
     */
    public void propertyChange(PropertyChangeEvent e) {
        String name = e.getPropertyName();
        if (name.startsWith("embeddedfile.")) {
            init((PDF) e.getSource());
        }
    }

    // -- ListModel --

    private class Model implements ListModel<EmbeddedFile> {

        public int getSize() {
            return model.size();
        }

        public EmbeddedFile getElementAt(int index) {
            return model.get(index);
        }

        public void addListDataListener(ListDataListener l) {
            synchronized (ldl) {
                ldl.add(l);
            }
        }

        public void removeListDataListener(ListDataListener l) {
            synchronized (ldl) {
                ldl.remove(l);
            }
        }

    }

    void fireContentsChanged() {
        ListDataEvent e = new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED, -1, -1);
        ListDataListener[] al;
        synchronized (ldl) {
            al = new ListDataListener[ldl.size()];
            ldl.toArray(al);
        }
        for (ListDataListener l : al) {
            l.contentsChanged(e);
        }
    }

    private void handleClick(EmbeddedFile attachment) {
        String type = attachment.getType();
        PDFViewer viewer = docpanel.getViewer();
        ViewerFeature[] features = viewer.getFeatures();
        for (ViewerFeature feature : features) {
            if (feature instanceof Importer) {
                Importer importer = (Importer) feature;
                if (importer.canLoad(type)) {
                    String title = attachment.getName();
                    InputStream in = attachment.getData();
                    importer.getImporter(viewer, in, title, null).start(this, UIManager.getString("PDFViewer.Loading"));
                    return;
                }
            }
        }
    }

    // -- ListCellRenderer --

    public Component getListCellRendererComponent(JList<? extends EmbeddedFile> list, EmbeddedFile attachment, int index, boolean isSelected, boolean hasFocus) {
        String name = attachment.getName();
        String description = attachment.getDescription();
        StringBuilder sb = new StringBuilder();
        sb.append("<b>" + name + "</b><br>");
        if (description != null) {
            sb.append(description);
        }

        JLabel comp = new JLabel("<html>"+sb+"</html>");
        comp.setBorder(BorderFactory.createEmptyBorder(0, 0, 10, 0));
        comp.setIcon(icon);
        comp.setOpaque(true);
        comp.setBackground(isSelected ? list.getSelectionBackground() : list.getBackground());
        comp.setForeground(isSelected ? list.getSelectionForeground() : list.getForeground());
        comp.setVerticalTextPosition(SwingConstants.TOP);
        return comp;
    }

}
